from mjooln import Doc, File, Key, FileName, Folder, Atom, \
    Element, FileSystem, Identity


class RootError(Exception):
    pass


class NotRootException(RootError):
    pass


class ErraticWakeSleep(RootError):
    pass


# TODO: Cleanup name Root in exceptions to reflect actual class name
class Root:
    """
    Combination of a folder and a file that defines a particular spot in
    the file system.

    .. warning:: Root is primarily meant to be used for inheritance, and
        not direct use

    Root key follows limitations of class :class:`.Key`.
    If root key is ``julian``,
    the following triplet of rules define a folder as a valid Root:

    1. Folder path is ``../julian``
    2. Folder contains a JSON file with name ``.julian.json``
    3. JSON file contains a dictionary, where one top level key is ``_root``,
       containing a :class:`.Root.RootDic` with ``Atom.key = 'julian'``

    This triplet defines a Root as valid independent of file system, database
    entry or configuration file entry, allowing it to be used as a standalone
    knot of data, identifiable by shape and not identity.

    The Root class also stores all attributes, except private, in the JSON
    file. The key ``_root`` is reserved for class private attributes.
    """
    # TODO: Add atom and species description. Also needs to be planted
    # TODO: Change key to name?
    # TODO: Allow compression and encryption

    #: Root class identifier
    ROOT = Element('root')

    #: Current class species (this is Root)
    SPECIES = ROOT

    # # TODO: Fix hack
    # _FILE_WILDCARD = '.*.json'

    # TODO: Make docfile that will update file when a variable is changed or
    # TODO: added
    TRUNK = Doc

    @classmethod
    def is_root(cls, folder):
        """
        Checks if folder is a valid root

        :param folder: Folder to check
        :type folder: Folder
        :return: True if folder is root, False if not
        :rtype: bool
        """
        try:
            _ = cls(folder)
            return True
        except NotRootException:
            return False

    @classmethod
    def _append_if_root(cls, roots, folder):
        if cls.is_root(folder):
            roots.append(cls(folder))

    @classmethod
    def walk_roots(cls, folder):
        folder = Folder.glass(folder)
        for fo in FileSystem.walk(folder,
                                  include_folders=True,
                                  include_files=False):
            if cls.is_root(fo):
                yield cls(fo)

    @classmethod
    def list_roots(cls,
                   folder,
                   pattern: str = '*',
                   recursive: bool = False):
        folder = Folder.glass(folder)
        return [cls(x) for x in
                FileSystem.list_folders(folder,
                                        pattern=pattern,
                                        recursive=recursive)
                if cls.is_root(x)]

    @classmethod
    def plant(cls,
              folder,
              key,
              trunk: TRUNK = TRUNK()):
        """
        Create a new root in folder, with given key and kwargs

        :param folder: Folder where the root will be created
        :type folder: Folder
        :param key: Root key, following :class:`.Key` limitations
        :type key: Key
        :param kwargs: Attributes to add to root
        :return: Created root
        :rtype: Root
        """
        # TODO: Require Ground to plant Root?
        folder = Folder.glass(folder)
        key = Key.glass(key)
        if not folder.exists():
            raise RootError(f'Folder does not exist: {folder}; '
                            f'Cannot plant root in non existent folder. '
                            f'Use Folder.create() first')
        root_folder = folder.append(key)
        if root_folder.exists():
            raise RootError(f'Cannot plant root in when folder by same name '
                            f'exists: {root_folder}. '
                            f'Empty and remove folder first. '
                            f'Or use a different key.')
        root_folder.create()
        cls._make_file(root_folder, key, trunk)
        return cls(root_folder)

    @classmethod
    def _make_file(cls,
                   root_folder,
                   key,
                   trunk: TRUNK = TRUNK()):
        root_folder = Folder.glass(root_folder)
        key = Key.glass(key)
        file = cls._get_file(root_folder)
        if file.exists():
            raise RootError(f'Cannot make existing {cls._name()} file: '
                            f'{file}')
        doc = {
            'atom': Atom(key).to_doc(),
            'species': cls.SPECIES.stub(),
            'trunk': trunk.to_doc(),
        }
        file.write(doc)

    @classmethod
    def _get_file(cls,
                  folder: Folder):
        file_name = FileName.make(stub=Key(folder.name()),
                                  extension=FileName.YAML,
                                  is_hidden=True)
        return File.join(folder, file_name)

    # TODO: Make exception texts more elfish
    @classmethod
    def elf(cls, folder):
        """
        Converts an existing folder to root, tolerating missing file but not
        invalid key. If a folder already is a valid root, it will be returned
        as root. The folder name must be a valid :class:`.Key`

        :raise RootError: If file already exists, but does not have the right
            key, or if it has an invalid format
        :param folder: Folder to convert to root
        :type folder: Folder
        :param kwargs: Attributes to add to root file
        :return: New or existing root
        :rtype: Root
        """
        if isinstance(folder, cls):
            return folder
        if isinstance(folder, str):
            folder = Folder(folder)
        if isinstance(folder, Folder) and cls.is_root(folder):
            return cls(folder)

        if not isinstance(folder, Folder):
            raise NotRootException(f'Cannot make {cls._name()} of this '
                                   f'input: {folder.__repr__()}')
        key = folder.name()
        if not Key.is_seed(key):
            raise NotRootException(f'Cannot make {cls._name()} when folder '
                                   f'name is not a valid Key: {folder.name()}')
        key = Key(key)
        if not folder.parent().exists():
            raise NotRootException(f'Cannot make {cls._name()} without a '
                                   f'parent folder. Create one first, and '
                                   f'use {cls._name()}.plant()')
        if not folder.exists():
            folder = folder.parent()
            return cls.plant(folder, key)

        file = cls._get_file(folder)

        if not file.exists():
            cls._make_file(folder, key, trunk=cls.TRUNK())
            # return cls(folder)

        return cls(folder)

    @classmethod
    def _name(cls):
        return cls.__name__

    def __init__(self, folder):
        self.__elf = False
        self._folder = None
        self._key = None
        self._file = None
        self.atom = None
        self.species = None
        self.trunk = None
        if folder is None:
            raise NotRootException(f'NoneType object is not a valid'
                                   f'{self._name()}. '
                                   f'Use {self._name()}.plant()')
        self._folder = Folder.glass(folder)
        if not self._folder.exists():
            raise NotRootException(f'Folder does not exist: '
                                   f'{self._folder}; '
                                   f'Use {self._name()}.plant()')
        if not Key.is_seed(self._folder.name()):
            raise NotRootException(f'Folder name \'{self._folder.name()}\' '
                                   f'is not a valid key, i.e. folder is '
                                   f'not a valid {self._name()}')
        self._key = Key(self._folder.name())
        self._file = self._get_file(folder)
        if not self._file.exists():
            raise NotRootException(f'File does not exist: {self._file}; '
                                   f'This is not a valid {self._name()}')
        self.wake()
        self.validate()

    def __str__(self):
        return self._folder.__str__()

    def __repr__(self):
        return f'Root(\'{self}\')'

    def validate(self):
        """
        Verify that root is valid

        :raise NotRootException: If root is not valid
        """
        if self._folder.name() != self._key:
            raise NotRootException('Folder key mismatch')
        if self.atom.key() != self._key:
            raise NotRootException('Atom key folder key mismatch')
        if self.species != self.SPECIES:
            raise NotRootException('Species mismatch')

    def _read(self):
        doc = self._file.read()
        if doc is None or \
                'atom' not in doc or \
                'species' not in doc or \
                'trunk' not in doc:
            raise NotRootException(f'Invalid contents for {self._name()} in '
                                   f'file: {self._file}')
        atom = Atom.from_doc(doc['atom'])
        species = Key(doc['species'])
        trunk = self.TRUNK.from_doc(doc['trunk'])
        if species != self.SPECIES:
            raise NotRootException(f'Species mismatch. '
                                   f'Species asleep: {species};'
                                   f'Species awake: {self.SPECIES}')
        return atom, species, trunk

    def _write(self):
        doc = {
            'atom': self.atom.to_doc(),
            'species': self.species.seed(),
            'trunk': self.trunk.to_doc(),
        }
        self._file.write(doc)

    def asleep(self) -> bool:
        return self.atom is None and \
               self.species is None and \
               self.trunk is None

    def awake(self) -> bool:
        return not self.asleep()

    def wake(self):
        if self.awake():
            raise RootError(f'Already awake. Use touch()')
        self.atom, self.species, self.trunk = self._read()

    def touch(self):
        if self.asleep():
            self.wake()
            return
        atom, species, trunk = self._read()
        if atom != self.atom or species != self.species:
            raise ErraticWakeSleep(
                f'{self._name()} is neither asleep nor awake. '
                f'Asleep atom: {atom}; '
                f'Awake atom: {self.atom};'
                f'Asleep species: {species};'
                f'Awake species: {self.species}')
        if trunk != self.trunk:
            self._write()

    def sleep(self):
        self.touch()
        self.atom = None
        self.species = None
        self.trunk = None


    def walk(self,
             include_folders: bool = True,
             include_files: bool = True):
        for x in FileSystem.walk(self._folder,
                                 include_folders=include_folders,
                                 include_files=include_files):
            if x != self._file:
                yield x

    def files(self):
        return self.walk(include_files=True,
                         include_folders=False)

    def folders(self):
        return self.walk(include_files=False,
                         include_folders=True)

    def list(self,
             pattern: str = '*',
             recursive: bool = False):
        return [x for x in
                FileSystem.list(self._folder,
                                pattern=pattern,
                                recursive=recursive)
                if x != self._file]

    def list_files(self, pattern='*', recursive=False):
        paths = self.list(pattern=pattern, recursive=recursive)
        return [x for x in paths if isinstance(x, File)]

    def list_folders(self, pattern='*', recursive=False):
        paths = self.list(pattern=pattern, recursive=recursive)
        return [x for x in paths if isinstance(x, Folder)]

    # TODO: Add choice to remove root, but leave contents?
    def uproot(self,
               with_force: bool = False,
               key: Key = None,
               identity: Identity = None):
        """
        Remove root. If root has contents, use ``with_force=True`` and
        ``key=<root key>`` and ``identity=<root identity>`` to force this

        .. warning:: Forcing uproot will remove all contents recursively

        :param with_force: Flags forced uproot, which will remove all files
            and folders recursively
        :param key: Add key to verify forced uproot
        """
        # TODO: Handle hidden files..
        if len(self.list()) > 0:
            if with_force:
                if key == self.atom.key() and identity == self.atom.identity():
                    self.sleep()
                    self._folder.empty(name=key)
                else:
                    raise RootError(f'Root folder ({self}) is not empty. '
                                    f'Enter root key and identity with '
                                    f'uproot_with_force=True: '
                                    f'key={self.atom.key()}, '
                                    f'identity={self.atom.identity()}')
            else:
                raise RootError(f'Root folder ({self}) is not empty. '
                                f'Uproot with force to empty it: '
                                f'with_force=True')
        if self._file.exists():
            self.sleep()
            self._file.delete()
        self._folder.remove()



import os
import base64

from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


class Crypt:
    """ Wrapper for best practice key generation and AES 128 encryption

    From `Fernet Docs <https://cryptography.io/en/latest/fernet/>`_:
    HMAC using SHA256 for authentication, and PKCS7 padding.
    Uses AES in CBC mode with a 128-bit key for encryption, and PKCS7 padding.
    """

    # TODO: Do QA on cryptographic strength

    @classmethod
    def generate_key(cls):
        """ Generates URL-safe base64-encoded random key with length 44 """
        return Fernet.generate_key()

    @classmethod
    def salt(cls):
        """ Generates URL-safe base64-encoded random string with length 24

        :return: bytes
        """

        # Used 18 instead of standard 16 since encode otherwise leaves
        # two trailing equal signs (==) in the resulting string
        return base64.urlsafe_b64encode(os.urandom(18))

    @classmethod
    def key_from_password(cls,
                          salt: bytes,
                          password: str):
        """ Generates URL-safe base64-encoded random string with length 44

        :type salt: bytes
        :type password: str
        :return: bytes
        """

        password = password.encode()  # Convert to type bytes
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=default_backend()
        )
        return base64.urlsafe_b64encode(kdf.derive(password))

    @classmethod
    def encrypt(cls,
                data: bytes,
                key: bytes):
        """ Encrypts input data with the given key

        :type data: bytes
        :type key: bytes
        :return: bytes
        """
        if key is None:
            raise CryptError(f'Encryption key missing, cannot encrypt')
        fernet = Fernet(key)
        return fernet.encrypt(data)

    @classmethod
    def decrypt(cls,
                data: bytes,
                key: bytes):
        """ Decrypts input data with the given key

        :type data: bytes
        :type key: bytes
        :return: bytes
        """
        if key is None:
            raise CryptError(f'Encryption key missing, cannot encrypt')
        fernet = Fernet(key)
        try:
            return fernet.decrypt(data)
        except InvalidToken as it:
            raise CryptError(f'Invalid token. Probably due to '
                             f'invalid password/key. Actual message: {it}')


class CryptError(Exception):
    pass




class Glass:
    @classmethod
    def glass(cls, *args, **kwargs):
        """
        If input is a class instance, return instance. If not, call
        constructor with same input arguments
        """
        if args and len(args) == 1 and not kwargs and isinstance(args[0], cls):
            return args[0]
        else:
            return cls(*args, **kwargs)


import logging
import datetime
import pytz
import dateutil
from dateutil.parser import parse as dateparser
import re
from collections import namedtuple

from mjooln.core.seed import Seed
from mjooln.core.glass import Glass

logger = logging.getLogger(__name__)


# TODO: Same date in all examples?
# TODO: Move examples to separate documentation page?


class ZuluError(Exception):
    pass


class Zulu(datetime.datetime, Seed, Glass):
    # TODO: Revise documentation
    # TODO: Round to millisecond etc. And floor. Check Arrow how its done
    # TODO: Move elf doc to elf method
    """Timezone aware datetime objects in UTC

    Create using constructor::

        Zulu() or Zulu.now()
            Zulu(2020, 5, 21, 20, 5, 31, 930343)

        Zulu(2020, 5, 12)
            Zulu(2020, 5, 12)

        Zulu(2020, 5, 21, 20, 5, 31)
            Zulu(2020, 5, 21, 20, 5, 31)

    Zulu.seed() is a string on the format ``<date>T<time>u<microseconds>Z``,
    and is \'designed\'
    to be file name and double click friendly, as well as easily recognizable
    within some string when using regular expressions.
    Printing a Zulu object returns seed, and Zulu can be created using
    from_seed()::

        z = Zulu(2020, 5, 12)
        print(z)
            20200512T000000u000000Z

        z.seed()
            '20200512T000000u000000Z'

        str(z)
            '20200512T000000u000000Z'

        Zulu.from_seed('20200512T000000u000000Z')
        z = Zulu(2020, 5, 12)

    For an ISO 8601 formatted string, use custom function::

        z = Zulu('20200521T202041u590718Z')
        z.iso()
            '2020-05-21T20:20:41.590718+00:00'

    Similarly, Zulu can be created from ISO string::

        Zulu.from_iso('2020-05-21T20:20:41.590718+00:00')
            Zulu(2020, 5, 21, 20, 20, 41, 590718)


    Inputs or constructors may vary, but Zulu objects are *always* UTC. Hence
    the name Zulu.

    Constructor also takes regular datetime objects, provided they have
    timezone info::

        dt = datetime.datetime(2020, 5, 23, tzinfo=pytz.utc)
        Zulu(dt)
            Zulu(2020, 5, 23, 0, 0, tzinfo=<UTC>)

        dt = datetime.datetime(2020, 5, 23, tzinfo=dateutil.tz.tzlocal())
        Zulu(dt)
            Zulu(2020, 5, 22, 22, 0, tzinfo=<UTC>)

    Zulu has element access like datetime, in addition to string convenience
    attributes::

        z = Zulu()
        print(z)
            20200522T190137u055918Z
        z.month
            5
        z.str.month
            '05'
        z.str.date
            '20200522'
        z.str.time
            '190137'

    Zulu has a delta convenience function for timedelta, as well as a function
    to add deltas directly to generate a new Zulu::

        Zulu.delta(hours=1)
            datetime.timedelta(seconds=3600)

        z = Zulu(2020, 1, 1)
        z.add(days=2)
            Zulu(2020, 1, 3)

    For more flexible ways to create a Zulu object, see :meth:`Zulu.elf`

    .. warning:: Elves are fickle and rude


    """

    _ZuluStr = namedtuple('ZuluStr', [
        'year',
        'month',
        'day',
        'hour',
        'minute',
        'second',
        'microsecond',
        'date',
        'time',
        'seed',
    ])

    _FORMAT = '%Y%m%dT%H%M%Su%fZ'
    REGEX = r'\d{8}T\d{6}u\d{6}Z'
    LENGTH = 23

    ISO_REGEX_STRING = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-' \
                       r'(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):' \
                       r'([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-]' \
                       r'(?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
    ISO_REGEX = re.compile(ISO_REGEX_STRING)

    ############################################################################
    # String methods
    ############################################################################

    @classmethod
    def is_iso(cls, str_):
        # TODO: Use regex instead, and include check on timezone (optional)
        """Check if input string is
        `ISO 8601 <https://en.wikipedia.org/wiki/ISO_8601>`_

        Check is done using regex ``Zulu.ISO_REGEX``

        :param ts_str: Maybe an ISO formatted string
        :return: True if str_ is iso, False if not
        :rtype: bool
        """
        try:
            if cls.ISO_REGEX.match(str_) is not None:
                return True
        except:
            pass
        return False

    ############################################################################
    # Timezone methods
    ############################################################################

    @classmethod
    def all_timezones(cls):
        """
        Returns a list of all allowed timezone names

        Timezone \'local\' will return a datetime object with local timezone,
        but is not included in this list

        Wrapper for ``pytz.all_timezones``

        :return: List of timezones
        :rtype: list
        """
        return pytz.all_timezones

    @classmethod
    def _to_utc(cls, ts):
        return ts.astimezone(pytz.utc)

    @classmethod
    def _tz_from_name(cls, tz='utc'):
        if tz == 'local':
            tz = dateutil.tz.tzlocal()
        else:
            try:
                tz = pytz.timezone(tz)
            except pytz.exceptions.UnknownTimeZoneError as ue:
                raise ZuluError(f'Unknown timezone: \'{tz}\'. '
                                f'Use Zulu.all_timezones() for a list '
                                f'of actual timezones')
        return tz

    ############################################################################
    # Create methods
    ############################################################################

    @classmethod
    def now(cls,
            tz=None):
        """
        Overrides ``datetime.datetime.now()``. Equivalent to ``Zulu()``

        :raise ZuluError: If parameter ``tz`` has a value. Even if value is UTC
        :param tz: Do not use. Zulu is always UTC
        :return: Zulu
        """
        if tz:
            raise ZuluError(f'Zulu.now() does not allow input time zone info. '
                            f'Zulu is always UTC. Hence the name')
        return cls()

    @classmethod
    def _from_unaware(cls, ts, tz=None):
        if not tz:
            raise ZuluError('No timezone info. Set timezone to use '
                            'with \'tz=<timezone string>\'. \'local\' will '
                            'use local timezone info. Use '
                            'Zulu.all_timezones() for a list of actual '
                            'timezones')
        return ts.replace(tzinfo=cls._tz_from_name(tz))

    @classmethod
    def _elf(cls, ts, tz=None):
        # Takes a datetime.datetime object and adds the input tzinfo if
        # none is present
        if not ts.tzinfo:
            ts = cls._from_unaware(ts, tz=tz)
        return ts

    @classmethod
    def from_unaware(cls, ts, tz='utc'):
        """ Create Zulu from timezone unaware datetime

        :param ts: Unaware time stamp
        :type ts: datetime.datetime
        :param tz: Time zone, with 'utc' as default.
            'local' will use local time zone
        :return: Zulu
        :rtype: Zulu
        """
        if ts.tzinfo:
            raise ZuluError(f'Input datetime already has '
                            f'time zone info: {ts}. '
                            f'Use constructor or Zulu.elf()')
        else:
            ts = cls._from_unaware(ts, tz=tz)
        return cls(ts)

    @classmethod
    def from_unaware_local(cls, ts):
        """ Create Zulu from timezone unaware local timestamp

        :param ts: Timezone unaware datetime
        :type ts: datetime.datetime
        :return: Zulu
        :rtype: Zulu
        """
        return cls.from_unaware(ts, tz='local')

    @classmethod
    def from_unaware_utc(cls, ts):
        """ Create Zulu from timezone unaware UTC timestamp

        :param ts: Timezone unaware datetime
        :type ts: datetime.datetime
        :return: Zulu
        :rtype: Zulu
        """
        return cls.from_unaware(ts, tz='utc')

    @classmethod
    def _parse_iso(cls,
                   iso: str):
        ts = dateparser(iso)
        if ts.tzinfo and str(ts.tzinfo) == 'tzutc()':
            ts = ts.astimezone(pytz.utc)
        return ts

    @classmethod
    def from_iso(cls,
                 str_: str,
                 tz=None):
        """
        Create Zulu object from ISO 8601 string

        :param iso: ISO 8601 string
        :param tz: Timezone string to use if missing in ts_str
        :return: Zulu
        :rtype: Zulu
        """
        ts = cls._parse_iso(str_)
        if tz and not ts.tzinfo:
            ts = cls._from_unaware(ts, tz)
        elif ts.tzinfo and tz:
            raise ZuluError(f'Timezone info found in ISO string as well as '
                            f'input timezone argument (tz). Keep tz=None, '
                            f'or use Zulu.elf()')
        elif not tz and not ts.tzinfo:
            raise ZuluError('No timezone info in neither ISO string '
                            'nor tz argument')
        return cls(ts)

    @classmethod
    def _parse(cls,
               ts_str: str,
               pattern: str):
        return datetime.datetime.strptime(ts_str, pattern)

    @classmethod
    def parse(cls,
              ts_str: str,
              pattern: str,
              tz=None):
        """Parse time stamp string with the given pattern

        :param ts_str: Timestamp string
        :type ts_str: str
        :param pattern: Follows standard
            `python strftime reference <https://strftime.org/>`_
        :param tz: Timezone to use if timestamp does not have timezone info
        :return: Zulu
        """
        ts = cls._parse(ts_str, pattern)
        if not ts.tzinfo:
            ts = cls._from_unaware(ts, tz=tz)
        elif tz:
            raise ZuluError('Cannot have an input timezone argument when '
                            'input string already has timezone information')
        return cls(ts)

    @classmethod
    def from_seed(cls, str_: str):
        """
        Create Zulu object from Zulu seed

        :param str_: Zulu seed
        :return: Zulu
        :rtype: Zulu
        """
        if not cls.is_seed(str_):
            raise ZuluError(f'String is not Zulu seed: {str_}')
        ts = cls._parse(str_, cls._FORMAT)
        ts = cls._from_unaware(ts, tz='utc')
        return cls(ts)

    @classmethod
    def _from_epoch(cls, epoch):
        ts = datetime.datetime.utcfromtimestamp(epoch)
        return ts.replace(tzinfo=pytz.UTC)

    @classmethod
    def from_epoch(cls, epoch):
        """
        Create Zulu object from UNIX Epoch

        :param epoch: Unix epoch
        :type epoch: float
        :return: Zulu
        :rtype: Zulu
        """
        ts = cls._from_epoch(epoch)
        return cls(ts)

    @classmethod
    def _fill_args(cls, args):
        if len(args) < 8:
            # From date
            args = list(args)
            args += (8 - len(args)) * [0]
            if args[1] == 0:
                args[1] = 1
            if args[2] == 0:
                args[2] = 1
            args = tuple(args)

        if args[-1] not in [None, 0, pytz.utc]:
            raise ZuluError(f'Zulu can only be UTC. '
                            f'Invalid timezone: {args[-1]}')

        args = list(args)
        args[-1] = pytz.utc
        return tuple(args)



    @classmethod
    def elf(cls, *args, tz='local'):
        """General input Zulu constructor

        .. warning:: Elves are fickle and rude

        ``Zulu.elf()`` takes the same inputs as constructor, and also allows Zulu
        objects to pass through. If timeozone is missing it will assume the input
        timezone ``tz``, which is set to local as default

        It takes both seed strings and iso strings::

            Zulu.elf('20201112T213732u993446Z')
                Zulu(2020, 11, 12, 21, 37, 32, 993446)

            Zulu.elf('2020-11-12T21:37:32.993446+00:00')
                Zulu(2020, 11, 12, 21, 37, 32, 993446)

        It takes UNIX epoch::

            e = Zulu(2020, 1, 1).epoch()
            e
                1577836800.0
            Zulu.elf(e)
                Zulu(2020, 1, 1)

        It will guess the missing values if input integers are not a full date
        and/or time::

            Zulu.elf(2020)
                Zulu(2020, 1, 1)

            Zulu.elf(2020, 2)
                Zulu(2020, 2, 1)

            Zulu.elf(2020,1,1,10)
                Zulu(2020, 1, 1, 10, 0, 0)

        .. warning:: Elves are fickle and rude

        :param args: Input arguments
        :param tz: Time zone to assume if missing. 'local' will use local
            time zone. Use Zulu.all_timezones() for a list of actual
            timezones. Default is 'local'
        :return: Zulu
        """
        ts = None
        if len(args) == 0:
            return cls()
        elif len(args) > 1:
            args = cls._fill_args(args)
            ts = datetime.datetime(*args)
            if not ts.tzinfo:
                ts = cls._from_unaware(ts, tz)
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, Zulu):
                return arg
            elif isinstance(arg, datetime.datetime):
                # Add timzone if missing
                ts = cls._elf(arg, tz=tz)
                return cls(ts)
            elif isinstance(arg, float):
                return cls.from_epoch(arg)
            elif isinstance(arg, int):
                # Instantiate as start of year
                return cls(arg, 1, 1)
            elif isinstance(arg, str):
                if cls.is_seed(arg):
                    return cls.from_seed(arg)
                elif cls.is_iso(arg):
                    ts = cls._parse_iso(arg)
                    # Add timzone if missing
                    ts = cls._elf(ts, tz=tz)
                else:
                    raise ZuluError(f'String is neither zulu, nor ISO: {arg}. '
                                    f'Use Zulu.parse() and enter the format '
                                    f'yourself')
            else:
                raise ZuluError(f'Found no way to interpret input '
                                f'argument as Zulu: {arg} [{type(arg)}]')
        return cls(ts)

    @classmethod
    def range(cls,
              start=None,
              n=10,
              delta=datetime.timedelta(hours=1)):
        """Generate a list of Zulu of fixed intervals

        .. warning:: Mainly for dev purposes. There are far better
            ways of creating a range of timestamps, such as using pandas.

        :param start: Start time Zulu, default is *now*
        :type start: Zulu
        :param n: Number of timestamps in range, with default 10
        :type n: int
        :param delta: Time delta between items, with default one hour
        :type delta: datetime.timedelta
        :return: [Zulu]
        """
        if not start:
            start = cls()
        return [Zulu.elf(start + x * delta) for x in range(n)]

    def __new__(cls, *args, **kwargs):
        if len(args) == 0 and len(kwargs) == 0:
            ts = datetime.datetime.utcnow()
            ts = ts.replace(tzinfo=pytz.UTC)
        elif len(args) == 1 and len(kwargs) == 0:
            arg = args[0]
            if isinstance(arg, str):
                raise ZuluError('Cannot instantiate Zulu with a string. Use '
                                'Zulu.from_iso(), Zulu.from_seed(), '
                                'Zulu.from_string() or Zulu.parse()')
            elif isinstance(arg, float):
                raise ZuluError(f'Cannot create Zulu object from a float: '
                                f'{arg}; If float is unix epoch, '
                                f'use Zulu.from_epoch()')
            elif isinstance(arg, Zulu):
                raise ZuluError(f'Input argument is already Zulu: {arg}. '
                                f'Use Zulu.elf() for a softer, but more '
                                f'fickle approach')
            elif isinstance(arg, datetime.datetime):
                ts = arg
                if not ts.tzinfo:
                    raise ZuluError('Cannot create Zulu from datetime if '
                                    'datetime object does not have timezone '
                                    'info. Use Zulu.from_unaware()')
                ts = ts.astimezone(pytz.UTC)
            else:
                raise ZuluError(f'Unable to interpret input argument: '
                                f'{arg} [{type(arg).__name__}]')
        else:
            # Handle input as regular datetime input (year, month, day etc)
            try:
                ts = datetime.datetime(*args)
            except TypeError as te:
                raise ZuluError from te
            # Add timezone info if missing (assume utc, of course)
            if not ts.tzinfo:
                ts = ts.replace(tzinfo=pytz.UTC)

        # Create actual object
        args = tuple([ts.year, ts.month, ts.day,
                      ts.hour, ts.minute, ts.second,
                      ts.microsecond, ts.tzinfo])
        self = super().__new__(cls, *args)
        seed = self.strftime(self._FORMAT)
        self.str = self._ZuluStr(
            year=seed[:4],
            month=seed[4:6],
            day=seed[6:8],
            hour=seed[9:11],
            minute=seed[11:13],
            second=seed[13:15],
            microsecond=seed[16:22],
            date=seed[:8],
            time=seed[9:15],
            seed=seed,
        )
        return self

    def __str__(self):
        return self.str.seed

    def __repr__(self):
        times = [self.hour, self.minute, self.second]
        has_micro = self.microsecond > 0
        has_time = sum(times) > 0
        nums = [self.year, self.month, self.day]
        if has_time or has_micro:
            nums += times
        if has_micro:
            nums += [self.microsecond]
        numstr = ', '.join([str(x) for x in nums])
        return f'Zulu({numstr})'

    def epoch(self):
        """
        Get UNIX epoch (seconds since January 1st 1970)

        Wrapper for ``datetime.datetime.timestamp()``

        :return: Seconds since January 1st 1970
        :rtype: float
        """
        return self.timestamp()

    @classmethod
    def from_str(cls, str_):
        if cls.is_seed(str_):
            return cls.from_seed(str_)
        elif cls.is_iso(str_):
            return cls.from_iso(str_)
        else:
            raise ZuluError(f'Unknown string format (neither seed nor iso): '
                            f'{str_}; '
                            f'Use Zulu.parse() to specify format pattern and '
                            f'timezone')

    def iso(self, full=False):
        # TODO: Implement full flag
        """Create `ISO 8601 <https://en.wikipedia.org/wiki/ISO_8601>`_ string

        Example::

            z = Zulu(2020, 5, 21)
            z.iso()
                '2020-05-21T00:00:00+00:00'

        :return: str
        """
        if full:
            raise ZuluError('Full isoformat not implemented. Full means it '
                            'has fixed length no matter the value. Is needed '
                            'for certain document database based tools')
        return self.isoformat()

    def format(self, pattern):
        """Format Zulu to string with the given pattern

        :param pattern: Follows standard
            `Python strftime reference <https://strftime.org/>`_
        :return: str
        """
        return self.strftime(pattern)

    def to_unaware(self):
        """
        Get timezone unaware datetime object in UTC

        :return: Timezone unaware datetime
        :rtype: datetime.datetime
        """
        return datetime.datetime.utcfromtimestamp(self.epoch())

    def to_tz(self, tz='local'):
        """ Create regular datetime with input timezone

        For a list of timezones use :meth:`.Zulu.all_timezones()`, which is a
        wrapper for

        :param tz: Time zone to use. 'local' will return the local time zone.
            Default is 'local'
        :return: datetime.datetime
        """
        ts_utc = datetime.datetime.utcfromtimestamp(self.epoch())
        ts_utc = ts_utc.replace(tzinfo=pytz.UTC)
        return ts_utc.astimezone(self._tz_from_name(tz))

    def to_local(self):
        """ Create regular datetime with local timezone

        :return: datetime.datetime
        """
        return self.to_tz(tz='local')

    @classmethod
    def delta(cls,
              days=0,
              hours=0,
              minutes=0,
              seconds=0,
              microseconds=0,
              weeks=0):
        """Wrapper for datetime.timedelta

        :param days: Number of days
        :param hours: Number of hours
        :param minutes: Number of minutes
        :param seconds: Number of seconds
        :param microseconds: Number of microseconds
        :param weeks: Number of weeks
        :return: datetime.timedelta
        """
        return datetime.timedelta(days=days,
                                  hours=hours,
                                  minutes=minutes,
                                  seconds=seconds,
                                  microseconds=microseconds,
                                  weeks=weeks)

    def add(self,
            days=0,
            hours=0,
            minutes=0,
            seconds=0,
            microseconds=0,
            weeks=0):
        """Adds the input to current Zulu object and returns a new one

        :param days: Number of days
        :param hours: Number of hours
        :param minutes: Number of minutes
        :param seconds: Number of seconds
        :param microseconds: Number of microseconds
        :param weeks: Number of weeks
        :return: Zulu
        """
        delta = self.delta(days=days,
                           hours=hours,
                           minutes=minutes,
                           seconds=seconds,
                           microseconds=microseconds,
                           weeks=weeks)
        return self + delta



import re


class BadSeed(Exception):
    pass


class Seed:
    """
    Convenience methods for string representation of an object

    Object can be created with the method ``from_seed()``, but the method
    must be overridden in child class. Find methods use the class variable
    ``REGEX``, which must also be overridden in child class

    If the seed has a fixed length, this can be specified in the class
    variable ``LENGTH``, and will speed up identification (or will it...)
    """

    #: Regex identifying seed must be overridden in child class
    REGEX = None

    #: If seed has a fixed length, override in child class
    LENGTH = None

    @classmethod
    def _search(cls, str_: str):
        if not cls.REGEX:
            raise BadSeed(f'_REGEX must be overridden in child class')
        return re.compile(cls.REGEX).search(str_)

    @classmethod
    def _exact_match(cls, str_: str):
        if not cls.REGEX:
            raise BadSeed(f'_REGEX must be overridden in child class')
        _regex_exact = rf'^{cls.REGEX}$'
        return re.compile(_regex_exact).match(str_)

    @classmethod
    def verify_seed(cls, str_: str):
        """
        Check if string is seed

        :raise BadSeed: If string is not seed
        :param str_: Seed to verify
        """
        if not cls.is_seed(str_):
            raise BadSeed(f'Sting is not seed: {str_}')

    @classmethod
    def is_seed(cls, str_: str):
        """
        Checks if input string is an exact match for seed

        :param str_: Input string
        :return: True if input string is seed, False if not
        """
        if cls.LENGTH and len(str_) != cls.LENGTH:
            return False
        return cls._exact_match(str_) is not None

    @classmethod
    def seed_in(cls, str_: str):
        """ Check if input string contains one or more seeds

        :param str_: String to check
        :type str_: str
        :return: True if input string contains one or more seeds, false if not
        """
        if cls._search(str_):
            return True
        else:
            return False

    @classmethod
    def find_seed(cls, str_: str):
        """
        Looks for and returns exactly one object from text

        Uses ``from_seed()`` to instantiate object from seed and will fail if
        there are none or multiple seeds.
        Use find_all() to return a list of identities in text, including
        an empty list if there are none

        :raise BadSeed: If none or multiple seeds are found in string
        :param str_: String to search for seed
        :type str_: str
        :return: Seed object
        """
        res = re.findall(cls.REGEX, str_)
        if len(res) == 1:
            return cls.from_seed(res[0])
        elif not res:
            raise BadSeed(
                f'No {cls.__name__} found in this text: \'{str_}\'; '
                f'Consider using find_seeds(), which will '
                f'return empty list if none are found.')
        else:
            raise BadSeed(
                f'Found {len(res)} instances of {cls.__name__} in this '
                f'text: \'{str_}\'; '
                f'Use find_all() to return a list of all instances'
            )

    @classmethod
    def find_seeds(cls,
                   str_: str):
        """ Finds and returns all seeds in text

        :type str_: str
        :return: List of objects
        """
        ids = re.findall(cls.REGEX, str_)
        return [cls.from_seed(x) for x in ids]

    @classmethod
    def from_seed(cls, str_: str):
        """
        Must be overridden in child class.

        Will create an object from seed

        :param str_: Seed
        :return: Instance of child class
        """
        raise BadSeed(f'Method from_seed() must be overridden '
                      f'in child class \'{cls.__name__}')

    def seed(self):
        """
        Get seed of current object.

        Default is ``str(self)``
        :return: Seed
        """
        return str(self)



import os
import glob
import logging

from mjooln.filesystem.folder import Folder, FolderError
from mjooln.filesystem.file import File, FileError
from mjooln.filesystem.path import Path, PathError

logger = logging.getLogger(__name__)


class FileSystemError(Exception):
    pass


class FileSystem:
    # TODO: Move factoryish methods in here from other path classes
    # TODO: Add convenience methods for Tree
    # TODO: Make separate folder for filesystem. Core has crypt and dicdoc
    # TODO: Allow none as input to liste etc, so that a default can be used (current dir for ex)
    # TODO: Change all to staticmethod

    @classmethod
    def folder(cls, *args):
        return Folder.join(*args)

    @classmethod
    def file(cls, *args):
        return File.join(*args)

    # @classmethod
    # def _walk(cls,
    #          folder: Folder,
    #          folder_handler=None,
    #          file_handler=None):
    #     for root, fos, fis in os.walk(folder):
    #         if folder_handler:
    #             for fo in (cls.folder(root, x) for x in fos):
    #                 folder_handler(fo)
    #         if file_handler:
    #             for fi in (cls.file(root, x) for x in fis):
    #                 file_handler(fi)

    @classmethod
    def walk(cls,
             folder,
             include_files: bool = True,
             include_folders: bool = True):
        folder = Folder.glass(folder)
        for root, fos, fis in os.walk(folder):
            if include_folders:
                for fo in (cls.folder(root, x) for x in fos):
                    yield fo
            if include_files:
                for fi in (cls.file(root, x) for x in fis):
                    yield fi

    @classmethod
    def files(cls, folder):
        folder = Folder.glass(folder)
        return cls.walk(folder,
                        include_files=True,
                        include_folders=False)

    @classmethod
    def folders(cls, folder):
        folder = Folder.glass(folder)
        return cls.walk(folder,
                        include_files=False,
                        include_folders=True)

    @classmethod
    def list(cls,
             folder,
             pattern='*',
             recursive=False):
        folder = Folder.glass(folder)
        paths = folder.list(pattern=pattern, recursive=recursive)
        for i, path in enumerate(paths):
            try:
                if path.is_file():
                    paths[i] = File(path)
                elif path.is_folder():
                    paths[i] = Folder(path)
            except FileError or PathError or FolderError:
                # TODO: Handle links and other exceptions
                pass
        return paths

    @classmethod
    def list_files(cls,
                   folder,
                   pattern='*',
                   recursive=False):
        folder = Folder.glass(folder)
        paths = cls.list(folder,
                         pattern=pattern,
                         recursive=recursive)
        return [x for x in paths if isinstance(x, File)]

    @classmethod
    def list_folders(cls,
                     folder,
                     pattern='*',
                     recursive=False):
        folder = Folder.glass(folder)
        paths = cls.list(folder,
                         pattern=pattern,
                         recursive=recursive)
        return [x for x in paths if isinstance(x, Folder)]

    @classmethod
    def home(cls):
        return Folder.home()

    @classmethod
    def current(cls):
        return Folder.current()

    @classmethod
    def human_size(cls, size_bytes: int):
        # 2**10 = 1024
        power = 2 ** 10
        n = 0
        size = size_bytes
        power_labels = {0: '', 1: 'k', 2: 'M', 3: 'G', 4: 'T'}
        while size > power:
            size /= power
            n += 1
        return size, power_labels[n] + 'B'

    @classmethod
    def bytes_to_human(cls,
                       size_bytes: int,
                       min_precision=5):
        value, unit = cls.human_size(size_bytes=size_bytes)
        len_int = len(str(int(value)))
        if len_int >= min_precision or unit == 'B':
            len_dec = 0
        else:
            len_dec = min_precision - len_int
        return f'{value:.{len_dec}f} {unit}'

    @classmethod
    def disk_usage(cls,
                   folder,
                   include_folders: bool = False,
                   include_files: bool = True):
        """
        Recursively determines disk usage of all contents in folder

        :param folder:
        :param include_folders:
        :param include_files:
        :raise FileSystemError: If folder does not exist
        :return: Disk usage of folder and subfolders
        :rtype: int
        """
        folder = Folder.glass(folder)
        if not folder.exists():
            raise FileSystemError(f'Cannot determine disk usage of '
                                  f'non existent folder: {folder}')
        size = 0
        for root, fos, fis in os.walk(folder):
            if include_folders:
                for fo in [cls.folder(root, x) for x in fos]:
                    size += fo.size()
            if include_files:
                for fi in [cls.file(root, x) for x in fis]:
                    size += fi.size()
        return size

    @classmethod
    def remove_empty_folders(cls, folder):
        folder = Folder.glass(folder)
        for root, folders, files in os.walk(folder):
            root = Folder(root)
            if root != folder and not folders and not files:
                root.remove()

    @classmethod
    def print(cls,
              folder,
              indent: int = 2,
              include_files: bool = True,
              include_folders: bool = True):
        # TODO: Redo with sequential walkthrough, and max_depth
        folder = Folder.glass(folder)
        for path in cls.walk(folder,
                             include_files=include_files,
                             include_folders=include_folders):
            level = path.replace(folder, '').count('/')
            path_indent = ' ' * indent * level
            path_tag = ''
            if path.is_folder():
                path_tag = '/'
            print(f'{path_indent}{path.name()}{path_tag}')



import logging
import string

from mjooln.core.name import Name
from mjooln.core.seed import Seed
from mjooln.core.glass import Glass

logger = logging.getLogger(__name__)


class InvalidElement(Exception):
    pass


class Element(str, Seed, Glass):
    """
    Defines element string with limitations

    - Minimum length is 2
    - Allowed characters are

        - Lower case ascii ``a-z``
        - Digits ``0-9``
        - Underscore ``_``
    - Underscore and digits can not be the first character
    - Underscore can not be the last character
    - Can not contain double underscore since it acts as separator for elements
      in :class:`.Key`

    Sample elements::

        'simple'
        'with_longer_name'
        'digit1'
        'longer_digit2'

    """
    REGEX = r'(?!.*__.*)[a-z0-9][a-z_0-9]*[a-z0-9]'

    #: Allowed characters
    ALLOWED_CHARACTERS = string.ascii_lowercase + string.digits + '_'

    #: Allowed first characters
    ALLOWED_STARTSWITH = string.ascii_lowercase + string.digits

    #: Allowed last characters
    ALLOWED_ENDSWITH = string.ascii_lowercase + string.digits

    @classmethod
    def is_seed(cls, str_: str):
        if len(str_) == 1:
            if Name.MINIMUM_ELEMENT_LENGTH > 1:
                return False
            else:
                return str_ in cls.ALLOWED_STARTSWITH
        else:
            return super().is_seed(str_)

    def __new__(cls,
                element: str):
        # TODO: Add list as input, creating key with separator
        cls.verify_element(element)
        self = super().__new__(cls, element)
        return self

    def __repr__(self):
        return f'Element(\'{self}\')'

    def stub(self):
        return self.__str__()

    @classmethod
    def from_stub(cls, stub: str):
        return cls(stub)

    @classmethod
    def verify_element(cls, str_: str):
        """
        Verify that string is a valid element

        :param str_: String to verify
        :return: True if str_ is valid element, False if not
        """
        if not len(str_) >= Name.MINIMUM_ELEMENT_LENGTH:
            raise InvalidElement(
                f'Element too short. Element \'{str_}\' has '
                f'length {len(str_)}, while minimum length '
                f'is {Name.MINIMUM_ELEMENT_LENGTH}')
        if not str_[0] in cls.ALLOWED_STARTSWITH:
            raise InvalidElement(f'Invalid startswith. Element \'{str_}\' '
                                 f'cannot start with \'{str_[0]}\'. '
                                 f'Allowed startswith characters are: '
                                 f'{cls.ALLOWED_STARTSWITH}')
        if not str_[-1] in cls.ALLOWED_ENDSWITH:
            raise InvalidElement(f'Invalid endswith. Element \'{str_}\' '
                                 f'cannot end with \'{str_[-1]}\'. '
                                 f'Allowed endswith characters are: '
                                 f'{cls.ALLOWED_ENDSWITH}')
        invalid_characters = [x for x in str_ if x not in
                              cls.ALLOWED_CHARACTERS]
        if len(invalid_characters) > 0:
            raise InvalidElement(
                f'Invalid character(s). Element \'{str_}\' cannot '
                f'contain any of {invalid_characters}. '
                f'Allowed characters are: '
                f'{cls.ALLOWED_CHARACTERS}')
        if Name.ELEMENT_SEPARATOR in str_:
            raise InvalidElement(
                f'Element contains element separator, which is '
                f'reserved for separating Elements in a Key.'
                f'Element \'{str_}\' cannot contain '
                f'\'{Name.CLASS_SEPARATOR}\'')


    # @classmethod
    # def elf(cls, element):
    #     """ Allows key class to pass through instead of throwing exception
    #
    #     :param element: Input element string or element class
    #     :type element: str or Element
    #     :return: Element
    #     """
    #     # TODO: Handle input and guess a conversion that would match criteria
    #     if isinstance(element, Element):
    #         return element
    #     else:
    #         return cls(element)



import logging

from mjooln.core.name import Name
from mjooln.core.seed import Seed
from mjooln.core.doc import Doc
from mjooln.core.glass import Glass

from mjooln.atom.zulu import Zulu
from mjooln.atom.key import Key
from mjooln.atom.identity import Identity

logger = logging.getLogger(__name__)


class AtomError(Exception):
    pass


class Atom(Doc, Seed, Glass):
    """
    Triplet identifier intended for objects and data sets alike

    Format: ``<zulu>___<key>___<identity>``

    :class:`.Zulu` represents t0 or creation time

    :class:`.Key` defines grouping of the contents

    :class:`.Identity` is a unique identifier for the contents

    Constructor initializes a valid atom, and will raise an ``AtomError``
    if a valid atom cannot be created based on input parameters.

    The constructor must as minimum have :class:`.Key` as input::

        a = Atom('some_key')
        a.key
            'some_key'
        a.zulu
            Zulu(2020, 5, 22, 13, 13, 18, 179169, tzinfo=<UTC>)
        a.identity
            '060AFBD5_D865_4974_8E37_FDD5C55E7CD8'
        str(a)
            '20200522T131318u179169Z___some_key___060AFBD5_D865_4974_8E37_FDD5C55E7CD8'

    ``

    """

    REGEX = r'\d{8}T\d{6}u\d{6}Z\_\_\_[a-z][a-z_0-9]*[a-z0-9]\_\_\_' \
            r'[0-9A-F]{8}\_[0-9A-F]{4}\_[0-9A-F]{4}\_[0-9A-F]{4}\_[0-9A-F]{12}'

    @classmethod
    def from_seed(cls,
                  str_: str):
        if not cls.is_seed(str_):
            raise AtomError(f'Invalid atom seed: {str_}')
        zulu, key, identity = str_.split(Name.CLASS_SEPARATOR)
        return cls(key=Key(key),
                   zulu=Zulu.from_seed(zulu),
                   identity=Identity(identity))

    @classmethod
    def elf(cls, *args, **kwargs):
        if len(args) == 1 and kwargs is None:
            arg = args[0]
            if isinstance(arg, Atom):
                return arg
            if isinstance(arg, Key):
                return cls(arg)
            elif isinstance(arg, str):
                if Key.is_seed(arg):
                    return cls(arg)
                elif cls.is_seed(arg):
                    return cls.from_seed(arg)
                else:
                    raise AtomError(f'This input string is nowhere near '
                                    f'what I need to create an Atom: {arg}')
            elif Key.is_seed(arg):
                return cls(arg)
            else:
                raise AtomError(f'How the fuck am I supposed to create an atom '
                                f'based on this ridiculous excuse for an '
                                f'input: {arg} [{type(arg)}]')
        elif len(args) == 0:
            if 'key' not in kwargs:
                raise AtomError(f'At the very least, give me a key to work '
                                f'on. You know, key as thoroughly described '
                                f'in class Key')
            key = Key.elf(kwargs['key'])
            identity = None
            zulu = None
            if 'identity' in kwargs:
                identity = Identity.elf(kwargs['identity'])
            if 'zulu' in kwargs:
                zulu = Zulu.elf(kwargs['zulu'])
            return cls(key,
                       zulu=zulu,
                       identity=identity)
        raise AtomError(f'This is rubbish. Cannot make any sense of this '
                        f'mindless junk of input: '
                        f'args={args}; kwargs={kwargs}')

    @classmethod
    def default(cls):
        raise AtomError('Atom does not have a default value')

    @classmethod
    def from_dict(cls,
                  data: dict):
        zulu = data.pop('zulu')
        key = data.pop('key')
        identity = data.pop('identity')
        atom = cls(key=key, zulu=zulu, identity=identity)
        atom.add(data)
        return atom

    # TODO: Make key, identity, zulu private, with getters (keep hashable)
    def __init__(self,
                 key,
                 zulu: Zulu = None,
                 identity: Identity = None):
        """ Initializes a valid atom

        :param args: One argument is treated as a atom string and parsed.
        Three arguments is treated as (zulu, key, identity), in that order.
        :param kwargs: key, zulu, identity. key is required, while zulu and
        identity are created if not supplied.
        """
        super().__init__()
        # TODO: Replace with Key.glass()?
        if isinstance(key, str):
            if Key.is_seed(key):
                key = Key(key)
            elif Atom.is_seed(key):
                raise AtomError(f'Cannot instantiate with Atom seed. '
                                f'Use Atom.from_seed()')
            else:
                raise AtomError(f'Invalid key: {key} [{type(key).__name__}]')

        if not isinstance(key, Key):
            raise AtomError(f'Invalid instance of key: {type(key)}')

        if not zulu:
            zulu = Zulu()
        if not identity:
            identity = Identity()

        self.__zulu = zulu
        self.__key = key
        self.__identity = identity

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return Name.CLASS_SEPARATOR.join([str(self.__zulu),
                                          str(self.__key),
                                          str(self.__identity)])

    def __repr__(self):
        return f'Atom(key=\'{self.__key}\', ' \
               f'zulu={self.__zulu.__repr__()}, ' \
               f'identity={self.__identity.__repr__()})'

    def __hash__(self):
        return hash((self.__zulu, self.__key, self.__identity))

    def __lt__(self, other):
        return (self.__zulu, self.__key, self.__identity) < \
               (other.__zulu, other.__key, other.__identity)

    def __gt__(self, other):
        return (self.__zulu, self.__key, self.__identity) > \
               (other.__zulu, other.__key, other.__identity)

    def key(self):
        return self.__key

    def zulu(self):
        return self.__zulu

    def identity(self):
        return self.__identity

    def to_dict(self,
                ignore_private: bool = True):
        return {
            'zulu': self.__zulu,
            'key': self.__key,
            'identity': self.__identity,
        }

    def to_doc(self,
               ignore_private: bool = True):
        # TODO: Replace stub with to_doc/from_doc (from Doc)
        return {
            'zulu': self.__zulu.iso(),
            'key': self.__key.seed(),
            'identity': self.__identity.seed(),
        }

    @classmethod
    def from_doc(cls, doc: dict):
        return cls(
            zulu=Zulu.from_iso(doc['zulu']),
            key=Key(doc['key']),
            identity=Identity(doc['identity']),
        )

    def with_separator(self,
                       separator: str):
        """ Atom string with custom separator

        :param separator: Custom separator
        :return: str
        """
        return separator.join(str(self).split(Name.CLASS_SEPARATOR))

    @classmethod
    def level_count(self,
                    level: int = None):
        if level is None or level < 0:
            return 1
        else:
            return level

    def levels(self,
               key_level: int = None,
               date_level: int = None,
               time_level: int = 0):
        """ Create list of levels representing the Atom

        Intended usage is for creating folder paths for atom files

        **None**: Full value

        **0**: No value

        **Negative**: Number of elements as string

        **Positive**: Number of elements as list

        Example::

            a = Atom(key='this__is__a_key')
            str(a)
                '20200522T181551u626184Z___this__is__a_key___F45A9C74_DC5D_48A6_9468_FC7E343EC554'
            a.levels()
                ['this__is__a_key', '20200522']
            a.levels(None, None, None)
                ['this__is__a_key', '20200522', '181551']
            a.levels(0, 0, 0)
                []
            a.levels(3, 2, 1)
                ['this', 'is', 'a_key', '2020', '05', '18']
            a.levels(-3, -2, -1)
                ['this__is__a_key', '202005', '18']

        :param key_level: Key levels
        :param date_level: Date levels
        :param time_level: Time levels
        :return: List of level strings
        :rtype: list
        """

        # TODO: Use this to make  level counter as well
        # TODO: Make this as classmethod
        def make(parts, level, sep=''):
            if level == 0:
                return []
            elif level > 0:
                return parts[:level]
            else:
                return [sep.join(parts[:-level])]

        if key_level is None:
            keys = [str(self.__key)]
        else:
            # TODO: Use key sep. At least as default
            keys = make(self.__key.elements(), key_level,
                        sep=Name.ELEMENT_SEPARATOR)

        zs = self.__zulu.str
        if date_level is None:
            dates = [zs.date]
        else:
            dates = make([zs.year, zs.month, zs.day], date_level)
        if time_level is None:
            times = [zs.time]
        else:
            times = make([zs.hour, zs.minute, zs.second], time_level)

        return keys + dates + times



from distutils.core import setup

VERSION = '0.5.0.32'

with open('README') as f:
    long_description = f.read()

setup(
    name='mjooln',
    packages=['mjooln',
              'mjooln.atom',
              'mjooln.core',
              'mjooln.filesystem',
              'mjooln.tree'],
    version=VERSION,
    license='MIT',
    description='Environmentally Friendly File Handling',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Vemund Halmø Aarstrand',
    author_email='vemundaa@gmail.com',
    url='https://mjooln.readthedocs.io/en/latest/',
    keywords=['JSON', 'PATH', 'FILE', 'FOLDER', 'FILE HANDLING',
              'ENCRYPTION', 'COMPRESSION', 'AES', 'GZIP', 'UUID', 'UTC',
              'MD5 CHECKSUM', 'DATA SCIENCE', 'DATA LAKE'],
    install_requires=[
        'python-dateutil',
        'pytz',
        'simplejson',
        'psutil',
        'cryptography',
        'pyyaml',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)
